load('profiler.js')
defaultPassword = "HELLO";

/** Grid usage

load('grid.js')

//start listening on port 8080
grid = new Grid(8080, password) 

//load a binary stream into the grid. It will be transferred on demand.
grid.resource('text','file://test.txt') 

//post a message into the grid and supply a callback function
grid.post('wordcount','file://test.txt', function(l) {writeln(l)}) 

//evaluate a function in parallel with different input variables
writeln(grid.evaluate(function(x,y){return x*y}, [ [2,5], [2,7], [456,.7] ] ))

//when grid.post(message, params, callback, opaque), the callback function looks like this:
//job has data, status, id
function callback(opaque, data, job)
{
 writeln(workername,"(",id,") processed ",message,"(",params,") in ",job.time," ms, returning",data)
}

*/

/** Grid data types

worker {io: stream, busy: bool, messages: array, info: text, job: object}
messages {source: worker, name: string, data: string}
job {name: string, data: string, alerts:array, tries: number, result: string, onComplete: function, callback: object}

The grid is defined as a set of workers and a master controller. The controller
provides data and instructions to the workers. Each worker advertises the names
of the messages it is willing to accept. So, on startup, each worker contacts a
controller, which updates its node database.

Messages to a worker:
STATUS
DATA name length\n...
ALERT something to print on the screen
RUN message_name length\nparameters
QUIT

Messages from a worker:
HELLO
IDLE
RUNNING messagename length\nparameters
ACCEPT compute,scale,exec
ALERT text to print to stdout
GET name
DATA name length\n...
RESULT status length\n...
BUSY can't accept that last job, send it somewhere else
ERROR can't run that job
SEND messagename length\nparameters
SENDTO target messagename length\nparameters
POST messagename length\nparameters
POSTTO target messagename length\nparameters

*/

/** Grid worker, minimal implementation in JS

let data = []
let grid = new Stream("tcp://127.0.0.1:8082")
grid.writeln('HELLO')
grid.writeln('ACCEPT foo,bar')
grid.writeln('IDLE worker ready!')

while (1)
{
  let line = grid.readln()
  
  if (!line) continue
  
  if (line == 'QUIT') break
      
  let info = line.split(' ')
  let message = info.shift()
  
  if(message == 'ALERT')
  {
    writeln(text)
  }
  else if (message == 'RUN')
  {
    let length = parseInt(info[0])
    let sender = parseInt(info[1] || 0)
    let params = this.master.read(length)
    grid.writeln('RUNNING ',name)
    
    if (name == 'foo') //no return value
    { 
      grid.writeln('RESULT 0 0')
    }
    else if (name == 'bar') //return value
    { 
      let result = params.length.toString()
      grid.writeln('RESULT 0 ',result.length)
      grid.write(result)
    }
  }
  else if (message =='STATUS')
  {
      grid.writeln('IDLE')
  }
  else if (message == 'DATA')
  {
    let name = info[0]
    let length = parseInt(info[1])
    data[name] = grid.read(length)
  }
  else
  {
    writeln("Unknown instruction ",line)
  }
}

*/

/** Grid worker, minimal implementation in python

import socket

def recvcmd(source):
    command = ''
    char = ' '
    while len(char) and ord(char) != 10:
      char = source.recv(1)
      command = command + char
    return string.strip(command)

def recvbytes(source,length)
    msg = ''
    while len(msg) < length:
      msg += source.recv(length-len(msg))
    return msg
    
data = {}
grid = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

grid.connect(("192.168.0.11",8082))
grid.send("HELLO\n")
grid.send("ACCEPT foo bar\n")

while 1:
  line = recvcmd(grid)
  
  if line == '':
    continue
     
  if line == 'QUIT':
    break
  
  message, text = line.split(' ',1)
  
  if message == 'ALERT':
    print text
    
  elif message == 'RUN':
    name, length, sender = text.split(' ')
    params = recvbytes(grid,int(length))
    grid.send('RUNNING %s 0\n' % name)
    
    if name == 'foo': #no return value
      grid.send('RESULT 0 0\n')
      
    elif name == 'bar': #return value
      answer = "%d" % len(params)
      grid.send('RESULT 0 %d\n' %len(answer))
      grid.send(answer)
      
  elif message == 'STATUS':
    grid.send('IDLE\n')
    
  elif message == 'DATA':
    name, length = text.split(' ',1)
    data[name] = recvbytes(grid,int(length))
    
  elif message == 'QUIT':
    break
    
  else:
    print "Unknown instruction %s" % line

*/
function debug() {}
//debug = writeln
//debugMessage = ''
//function debug() {for (let i=0; i<arguments.length; i++) debugMessage+= arguments[i]; debugMessage += '\n';}
//if (system.arguments && system.arguments.indexOf('-debug') != -1)
// debug = writeln

function nextWord(line)
{
 let a = line.indexOf(' ') 
 if (a == -1) return [line,'']
 return ([line.substr(0,a),line.substr(a+1)])
}

function Grid(port,password,controlpw)
{
 this.running = false
 this.port = this.port || 8082
 this.password = password || defaultPassword;
 this.server = new Server(this.port)
 this.nextid = 1;
 this.busy = 0
 this.maxTries = 4;
 this.workers = []; // {id, io, name, buffer, busy, messages, info, job}
 this.id = 0; //don't confuse this with a worker
 this.queue = [] // outgoing jobs to workers, waiting to be done {worker, name, data, tries, onComplete, opaque}
                   // set worker > 0 to send a message to a specific node
 this.data = {} // shared binary resources name:stream
 this.files = {} // files shared with workers name:stream
 this.errors = []
 this.httpd = null
 this.controlpw = controlpw || "" // external connection for scheduling jobs

 this.timeout = 1000; 
 this.keepalive = new Date
}

Grid.prototype.toStream = function() {return this.server}

Grid.prototype.Worker = function(id, name, io)
{
 debug('Accepted ',id,' from ', io)
 this.id = id
 this.name = name
 this.time = new Date
 this.io = io
 this.buffer = ''
 this.messages = []
 this.busy = false 
 this.info = ''
 this.alert = ''
 this.command = ''
 this.job = null
 
 this.profiler = new Profiler('w'+id)
}

Grid.prototype.Worker.prototype.toStream = function() {return this.io}

Grid.prototype.Worker.prototype.read = function()
{
    if (this.io.canRead) this.command = this.io.readLine();
    return this.command.length != 0;
    /*
    let c = ''
    if (this.io !== undefined) 
    if (this.io)
    while (this.io.canRead)
    {
        c = this.io.read(1)
        if (!c.length) return false
        if (c == '\r') continue
        if (c == '\n') return true //(this.command.length > 0)
        else this.command += c;
        
        if (this.command.length > 1024) //buffer overrun
        {
         writeln('!Overrun from ', this.io)
         this.io.close()
         delete this.io
         break
        }
    }
    return false;
    */
}

Grid.prototype.Worker.prototype.readData = function(length)
{
 let c = ''
 if (!length || length == '0') return c
 c = this.io.read(Number(length))
 debug("DATA ",c)
 return c
}

Grid.prototype.Worker.prototype.reset = function()
{
    this.command = ''
}

Grid.prototype.Job = function(worker,name,data,onComplete,opaque)
{
 this.worker = worker
 this.name = name
 this.data = data
 this.tries = 0
 this.result = null
 this.onComplete = onComplete
 this.opaque = opaque
 this.alerts = []
 this.children = []
 this.time = null
 this.sender = null //worker who originiated the request
}


/* 
  Closes workers 
*/
Grid.prototype.quit = function(list)
{  
debug("Quitting")
  this.running = false;
  for (let i=0; i<this.workers.length; i++) 
  { let w = this.workers[i];
    if (list == null || list.indexOf(w.id) != -1)
    {
      [w] = this.workers.splice(i,1);
      if (w.io)
      {
          w.io.writeln('QUIT')
          //w.io.close()      
      }
      i--;
    }
  }    
  system.sleep(500)
}

Grid.prototype.find = function(id)
{
  for (let i=0;i< this.workers.length; i++)
    if (this.workers[i].id == id) return this.workers[i];
  return null;
}

Grid.prototype.match = function(id, message)
{
  for (let i=0;i< this.workers.length; i++)
  {let w=this.workers[i];
    if (w.id == id || (message && (w.name == id && w.messages.indexOf(message) != -1))) return w
  }
  return null
}

/* Add a job to the post */
Grid.prototype.post = function(name, params, onComplete, opaque)
{
 return this.postToHost(0,name,params,onComplete,opaque)
}

/* Add a job to the front of the post */
Grid.prototype.send = function(name, params, onComplete, opaque)
{
 return this.sendToHost(0,name,params,onComplete,opaque)
}

/* 
  Send a message to a set of workers if they accept that message
*/
Grid.prototype.postToList = function(ids, name, params, onComplete, opaque)
{
 if (!name) return
 
 let workers = ids.map(Grid.prototype.find,this)
 for each(let w in workers) 
  if (w && w.id && w.messages.indexOf(name) != -1)
  { 
   this.queue.push(new this.Job(w.id,name,params,onComplete,opaque) )
   this.busy ++
  }
}

/* 
  Send a message to all workers who accept that message
*/
Grid.prototype.postToAll = function(name, params, onComplete, opaque)
{
 var count = 0
 for (let i=0;i< this.workers.length; i++)
  {let w=this.workers[i];
   if (w && w.messages.indexOf(name) != -1)
  {
   this.queue.push(new this.Job(w.id,name,params,onComplete,opaque))
   this.busy ++
   count++
 } }
 return count
}

/* 
  Send a message to a all workers on a particular if they accept that message
*/
Grid.prototype.postToHostAll = function(host, name, params, onComplete, opaque)
{
 if (!name) return
 
 for (let i=0;i< this.workers.length; i++)
  {let w=this.workers[i];
   if (w && w.name == host && w.messages.indexOf(name) != -1)
  { 
   this.queue.push(new this.Job(w.id,name,params,onComplete,opaque)) 
   this.busy ++
 } }
}

/* 
  Send a message to any worker on a particular host who accepts that message  
*/
Grid.prototype.postToHost = function(host, name, params, onComplete, opaque)
{
 if (!name) return null
 if (host)
 {
  let w = this.match(host,name)
  if (!w) return null //if (!w || w.messages.indexOf(name) == -1) return null
 }
 let job = new this.Job(host,name,params,onComplete,opaque)
 this.queue.push(job) 
 this.busy ++
 return job
}

/* 
  Send a message to any worker on a particular host who accepts that message  
  And do it right away
*/
Grid.prototype.sendToHost = function(host, name, params, onComplete, opaque)
{
 if (!name) return null
 if (host)
 {
  let w = this.match(host, name)
  if (!w) return null //if (!w || w.messages.indexOf(name) == -1) return null
 }
 let job = new this.Job(host,name,params,onComplete,opaque)
 this.queue.unshift(job) 
 this.busy ++
 return job
}
/*
  Return a list of ids of workers who accept a specific message
*/
Grid.prototype.accepting = function(message)
{
 let ret = []
 for (let i=0;i< this.workers.length; i++)
  {let w=this.workers[i];
   if (w && w.messages.indexOf(message) != -1)
  {
   ret.push(w.id)
 } }
  return ret;
}

/* Add data to the grid */
Grid.prototype.resource = function(name)
{
  let s = this.data[name]
  if (s) s.rewind()
  else if (this.files[name]) 
  {
    s = this.files[name]
    s.rewind()
  }
  else if (system.files().indexOf(name) != -1)
  {
    //writeln('opening ',name)
    s = new Stream('file://' + name,'rb')
    this.files[name] = s
   }
  return s
}

/*
  Resources are cached and made available to workers who know their names.
  Single-read streams like web addresses should be converted to strings (with Stream.readFile())
  so that they can be used more than once.

  Note that files in the current directory are always available to workers

*/
Grid.prototype.addResource = function(name, data)
{
try{
 if (data.constructor.name == "Stream")
 {
  let s = new Stream
  s.append(data)
  this.data[name] = s
 }
 else if(typeof data == "string")
 {
  let s = new Stream
  s.write(data)
  this.data[name] = s
 }
} 
 catch(x) 
 {
  writeln("Add resource error:",x)
  for (let [n,v] in x) writeln(n, ': ',v)
  return false
 }
}

/*
  Assigns the next job from the post
*/
Grid.prototype.delegate = function(worker)
{
 if (worker.busy) return false;
try {
 for (let i=0; i<this.queue.length; i++)
 {
  let job = this.queue[i] 
  //writeln(job.name,' ',job.data)
  if (job.worker && job.worker != worker.id && job.worker != worker.name) 
    continue
  
  if (!worker.busy && worker.messages.indexOf(job.name) != -1)
  {
   worker.busy = true
   let d = (job.data || '').toString()
   debug(worker.id, '>RUN ', job.name,' ',d.length,' ',job.data)
   if (job.sender)
    worker.io.writeln('RUN ',job.name,' ',d.length,' ',job.sender)
   else
    worker.io.writeln('RUN ',job.name,' ',d.length)
   worker.io.write(d)
   job.time = new Date
   worker.job = job
   this.queue.splice(i,1)
   return true
  }
 }
 } catch(x)
 {
  writeln("Delegate error: ",x)
  for (let [n,v] in x) writeln(n, ': ',v)
 }
 return false
}

/*
Move messages that aren't going to be dispatched to the error post
*/
Grid.prototype.orphan = function()
{
  let c = 0
  for (let i=0; i< this.queue.length; i++)
  {
    if (this.queue[i].worker && this.find(this.queue[i].worker) == null)
    {
     this.errors.push(this.queue[i])
     this.queue.splice(i,1)
     this.busy--
     i--
     c++
    }
  }    
  return c
}

/** Randomize the workers for load balancing
*/
Grid.prototype.shuffle = function()
{
 let x = new Array(l)
 while  (this.workers.length)
 {
  x.push(this.workers.splice(Math.floor(Math.random() * this.workers.length),1)[0])
 }
 this.workers = x
}

/** 
  Runs the grid message loop, invoking an optional callback
*/
Grid.prototype.onSignal = function() {return this.accept(this.server.accept());};
Grid.prototype.Worker.prototype.onSignal = function(grid) {return grid.handleWorker(this);};

var p1 = new Profiler('p1')
//grid event loop
Grid.prototype.run = function(callback,httpd)
{
  this.running = true
  //this.shuffle()
  this.httpd = httpd
  let currentError = 0;
  while (this.running && (this.httpd || this.busy) && (!callback || callback(this)))
  {
    var any = false;
    //if (this.server.anyoneWaiting)
    //{
    //    any = true
    //    this.accept(this.server.accept())
    //    continue
    //}
    
    this.workers.push(this);
    var waiting = system.wait(this.workers,1000);
    //if (waiting) waiting.forEach(function(x) write(x.id,' '))
    this.workers.pop();
    if (waiting) 
    for (let i=0;i<waiting.length; i++)
      any |= waiting[i].onSignal(this);

    for (let i = 0; i < this.workers.length; i++)
      any |= this.delegate(this.workers[i]);
      
    if (!any)
    {
       this.idle()
       system.gc()
    }  

    for (; currentError < this.errors.length; currentError++)
    {
      let e = this.errors[currentError]
      let fields = ["worker","name","data","tries","result"]
      writeln('Error: {',fields.map(function(f) f+":"+e[f].toSource()),'}')
    }
   }   
 return !this.busy
}  

Grid.prototype.accept = function(connection)
{
 let line = connection.readln();
debug("accept ",line);

 let [name,pwd,protocol] = line.split(' ');
 if (this.httpd && protocol && protocol.indexOf('HTTP/') == 0)
 { 
  this.onHTTP(connection,line);
  connection.close();
 }
 else if (pwd == this.password)
 {
  let w = new this.Worker(this.nextid++, name, connection);
  debug('Adding ',w.id,':',connection,' ',name,' ',pwd);
  
  this.workers.push(w);
  return true;
 }
 else 
  connection.close();
 return false;
}

Grid.prototype.onHTTP = function(stream,request)
{
  if (!this.httpd.handleRequest(stream,request))
   this.running = false
  if (!this.httpd.running)
   this.running = false
}

/** 
   Are any new workers waiting to join the grid?
   Ping all the workers we haven't heard from recently. 
   If a worker is idle, give it the next job that it can accept from the post
*/
var p2 = new Profiler('p2')
var p3 = new Profiler('p3')
Grid.prototype.handleWorker = function(w)
{
debug("handle worker ",w.id,' ',w.io.eof,' ',w.io.canRead)
  if (w.io.eof && !w.io.canRead) 
    {
        this.workers.splice(this.workers.indexOf(w),1);
        return true;
    }
    
    let any=false;
    if (!w.io || !(w.io.canWrite || w.io.canRead)) 
    {
        if (w.job) 
        {
            w.job.tries++
            w.job.result = "Connection lost"
            if (w.tries > this.maxTries || w.job.worker) //can't reschedule a message intended for a specific recipient
            {
                this.errors.push(w.job)
                this.busy --
            }
            else
              this.queue.unshift(w.job)
        }
        if (w.io)
        debug(w.id,'!CLOSE ',w.io.name, ' ',w.io.canWrite,' ',w.io.canRead)
        if (w.io) w.io.close()
        w.io = null
        this.workers.splice(j,1)
        i--       
    }    
    else
    {
       try
        {
            p2.inc()
            
            any = this.listen(w) //any conversations?
        } catch(x)
        {
            w.io.close()
            w.io = null
            debug(w.id,'!ERR ', x)
            writeln("Communication error: ",x)
            if (typeof x != "string") for (let [n,v] in x) writeln(n, ': ',v)
        }  
    }
    return any;
}

Grid.prototype.idle = function()
{
  let now = new Date
  if (now - this.keepalive > 1000) //heartbeat every second
  {
      for (let i = 0; i < this.workers.length; i++)
       {
        if (this.workers[i].io) this.workers[i].io.write('\n') //hello
       }
      //system.sleep(100)
      this.keepalive=now
  }
 this.orphan(); //scrub the pending jobs list
 return (this.busy > 0 && this.workers.length > 0)  
}


/* 
  Read a reply from a connection and update states as needed 
*/
Grid.prototype.listen = function(w)
{
 let any = false
 
 if (w.read()) //true) 
 {
   let data = null
   //debug(w.id,',',w.io.canWrite,',',w.io.eof)
   let line = w.command //this.recvcmd(w)
   writeln(w.io,': ',line)
   //let line = w.io.readln() //this.recvcmd(w)
   if (!line) return false
   w.reset()
   any = true
   //debug(w.id,'>',line)
   let [message, text] = nextWord(line)
   message = message.toUpperCase()
   switch (message)
   {
     case 'IDLE': 
      w.info = 'IDLE '+text
      w.busy = false
      break
     case 'RUNNING': //RUNNING program_name length\n[text]
      {
          let [name,length] = text.split(' ')
          let data = w.readData(length)
          w.info = 'RUNNING ' + name
          if (data) w.info += ': ' + data
      }
      break
     case 'SEND': //SEND program_name length\n[text]
     case 'POST': //POST program_name length\n[text]
      {
          let [name,length] = text.split(' ')
          let data = w.readData(parseInt(length))
          debug(data)
          let job = null
          if (message == 'SEND')
           job = this.send(name,data).sender=w.id
          else
           job = this.post(name,data).sender=w.id
          
          if (job && w.job) w.job.children.push(job.id)
      }      
      break
     case 'SENDTO': //SENDTO target program_name length\n[text]
     case 'POSTTO': //POST program_name length\n[text]
      {
          let [host,name,length] = text.split(' ')
          let data = w.readData(parseInt(length))
          debug(data)
          let job = null
          if (message == 'SENDTO')
           job = this.sendToHost(parseInt(host),name,data).sender=w.id
          else
           job = this.postToHost(parseInt(host),name,data).sender=w.id
          
          if (job && w.job) w.job.children.push(job.id)
          else this.errors.push({host:host, name: name, data: data})
      }      
      break
     case 'ERROR':
      w.busy = false
      w.info = 'ERROR ' + text
      if (w.job) 
      {
        w.job.tries++
        w.job.result = text     
        this.busy --
        this.errors.push(w.job)
        w.job = null
      }
      break
     case 'BUSY':
     if (w.job) //got lost -- retry
      {
        this.queue.unshift(w.job)
        debug('**Job lost. busy: ',this.busy)
        w.job = null
      }
      w.info = 'BUSY ' + text
      break
     case 'RESULT': //RESULT time length\n[data]
      {
      let [status,length] = text.split(' ')
      let data = w.readData(parseInt(length))
      let job = w.job
      w.info = 'RESULT '+data
      w.history = job
      w.job = null //done!
      w.busy = false
      if (job) 
          {
            job.time = ( (new Date)-job.time) //elapsed
            job.status = status
            job.result = data
            job.worker = w.name
            job.id = w.id
            if (job.onComplete)
            {
                if (typeof job.opaque != "undefined")
                    job.onComplete(job.opaque, data, job)
                else                
                    job.onComplete(data)
            }
          }
          
      w.profiler.begin()
      this.busy --
      }
      break
     case 'ACCEPT': //ACCEPT program1,program2,...
      w.messages = text.split(/[,;\s]+/g)
      debug(w.id,':',w.messages.toSource())
      break
     case 'GET': //GET data_file_name, reply with DATA data_file_name length
      {
      debug("get:", text)
        let s = this.resource(text)       
        if (!s)
          w.io.writeln('DATA ',text,' 0')
        else
        {
         w.io.writeln('DATA ',text,' ',s.size)
         w.io.append(s)
        }
      }
      break
     case 'ALERT': //ALERT length text
      w.alert = w.io.read(Number(text))
      debug(w.id,': ',w.alert)
      if (w.job) w.job.alerts.push(w.alert)
      break      
     case 'DATA': //DATA name length
     {
      let [name,length] = text.split(' ')
      let s = new Stream
      s.append(w.io,parseInt(length))
      debug("put:",s)
      this.data[name] = s
     } break
     case 'BYE':
      w.io.writeln('BYE');
      break;
    default:
      debug(w.id,'>',line)
   }
 }
 return any
}

Grid.decompileFunction = function(fn)
{
 if (typeof fn == "string")
   return {name:"anonymous", params:[], code:fn, data:[]}
 //return ["anonymous","i",fn]   
   
 //let ret = fn.toString().match(/function (\w*)\(([^)]*)\) {\n(.*)\n}/m)
 //ret.shift()
 //return ret
 
 let [,name,params,body] = fn.toString().match(/^function (\w*)\(([^)]*)\) {\n([\s\S]*)\n}$/)
 return {name: name, params: params.split(/,\s*/), code: body, data: []}
}
/*
Evaluate a function for multiple input conditions
 grid.evaluate(function(x,y){return x*y}, [[5,6],[1,4],[6,2],[5,4],[2,3]] )
 returns [30,4,12,20,6]
 
RUN evaluate [length]\n
{params: ["a","b"], code: "return a*b", data:[1,5]}
*/

Grid.prototype.evaluate = function(fn, data)
{
//function.toString() -> "function name(params) {\nbody\n}"
 //let [name,params,body] = Grid.decompileFunction(fn)
 //let msg = {params: params.split(/,\s*/), code: body, data: []}
 let msg = Grid.decompileFunction(fn)
 let result = new Array(data.length)
 let handler = function (index,answer) {debug( 'Result[',index,']=',answer); result[index]=answer}
 for (let i = 0; i < data.length; i++)
 {
  msg.data = data[i]
  this.post("evaluate", msg.toSource(),handler, i)
 }
 this.run()
 return result
}

/*
Shortcut for generating an array of a range of integers to pass to evaluate()
*/
Grid.range = function(x,y)
{
 let r = new Array(1+Math.abs(y-x))
 for (let i=0; i<r.length; i++) r[i] = [(x + (i * (y-x) / (r.length-1)))]
 return r
}


/*
Evaluate many bits of code in parallel
 grid.parallel("return 5*6","return (new Stream('http://www.google.com')).readFile()")
*/

Grid.prototype.parallel = function() //function, data, function, data, ...
{
 let result = new Array(arguments.length)
 let handler = function (index,answer) {debug('Result[',index,']=',answer); result[index]=answer}
 msg = {params: [], code: [], data: []}
 for (let i = 0; i < arguments.length; i++)
 {
  let s = arguments[i]
  //if (s.indexOf('return') == -1) s = 'return ' + s
  if (typeof s == 'function')
      msg.code = Grid.decompileFunction(s)[2]
  else 
      msg.code = s.toString()
  this.post("evaluate", msg.toSource(),handler, i)
 }
 this.run()
 return result
}

Grid.test = function()
{
  let count = 3

  let grid = new Grid()  
  
  writeln("Grid running on ",grid.server,". Starting ", count, " nodes.")
  
  let cmd = 'gridworker.js '+this.server.toString()+ ' ' + this.password

  for (let i=0; i<count; i++)
   system.execute(system.program,cmd)
   
  writeln("\nEvaluate a function over many inputs in parallel")
  
  let points = []
  for (let i=1; i<=10; i++) for (let j=1; j<=10; j++) points.push([i,j])
  writeln(grid.evaluate(function(x,y){return x*y}, points ))

  writeln("\nEvaluate separate pieces of code in parallel")
  writeln(grid.parallel("return 5*6","return Math.sin(1.4)","return this.grid.resource('test.txt').readFile().length"))
  
  writeln("\nDispatch computations directly to the grid")
  // low-level interface: post up messages and individual response handlers
  //simple response handler. The "run" message captures the program output
  grid.post('run','perfect.js', write);
  
  //advanced response handler
  
  function handler(filename, words, time)
  {
   writeln(filename,"\t has ", words," words in "," in ",time," ms")
  }
  
  for each (let f in system.files())
   grid.post('wordcount',f,handler,f)
  
  //wait until pending operations have completed, allow a keyboard hit to stop it
  grid.run(function(){return !system.kbhit()})
  
  writeln('Done')

  //shut down workers  
  grid.quit()
}

function spin(label)
{
  let symbols = '-*/\|'
  write(label,symbols[Math.floor(Math.random()*symbols.length)],'\r')
}

function Worker(address,password,name)
{
 this.address = address || (system.name + ':8082')
 this.password = password || defaultPassword
 this.name = name || system.name
 this.master = null 
 this.busy = false
 this.job = ''
 this.running = true
 this.parameters = ''
 this.data = {}
 this.command = ''
 this.sender = 0 //who sent the message I'm currently processing?
 //this.waiter = 0
 this.handlers = {  "run": Worker.runScript, 
                    "load": Worker.load,
                    "stop": Worker.stop
 };
 
 let i=0
/* while (this.running && !this.connect(true)) //wait for a connection
  {
    system.sleep(100 * (i++))
    if (i > 10) i = 1;
  }*/
}

//block
Worker.prototype.read = function()
{
    let c = ''
    if (this.master) 
    while (1) //this.master.canRead)
    {
        c = this.master.read(1)     
        if (!c.length) return false //stalled connection
        if (c == '\r') continue
        if (c == '\n') return true //(this.command.length > 0)
        else this.command += c;
    }
    return false;
}

Worker.prototype.reset = function()
{
    this.command = ''
}
 
/*
RUN wordcount [length]\nfilename
*/
Worker.wordcount = function(filename,grid)
{              
 let text = grid.resource(filename).toString()
 return text.split(/\s+/g).length
}

Worker.stop = function(why,grid)
{
  writeln('Stop!',why);
  grid.running = false;
}
/*
RUN load [length]\nfilename.js
The file is copied to the local directory and then loaded into the runtime environment.
*/
Worker.load = function(name,grid)
{
 let filename = name.replace(/[:\/\\]/g,'_')
 let s = new Stream(filename,'wb')
 grid.resource(name,s)
 s.close()
 load(filename)
}
 
/*
RUN run [length]\nfilename.js
The file is copied to memory and then run (without saving on the local disk)
*/
Worker.runScript = function(script,grid)  
{
    let result = ''
    let writeln = function() {for (let i=0; i<arguments.length; i++) result += arguments[i]; result += '\n'}
    let write = function() {for (let i=0; i<arguments.length; i++) result += arguments[i]}
    eval(grid.resource(script).readFile()) 
    return result
}

/*
RUN evaluate [length]\n
{params: ["a","b"], code: "return a*b", data:[1,5]}}

or

some_function_call()
*/
Worker.evaluate = function(text,grid)
{
  var msg
  if (text[0] != '{' && (text[0] != '(' || text[1] != '{')) 
  {
    msg = {params: [text], data:[]}
  }
  else
  {
    msg = eval('('+text+')')  
    msg.params.push(msg.code) //"a","b","return a*b"
  }
  let f = Function.constructor.apply(Function, msg.params)
  f = f.apply({grid:grid},msg.data) || 0;
  return f.toSource()
}

Worker.prototype.connect = function()
{ 
 if (this.master)
 {  
  //writeln(this.master.canWrite,this.master.canRead,this.master.eof)
  if (!this.master.canRead && this.master.eof) 
  {
   this.master.close();
   this.master = null;
  }
  //this.master.writeln()
  //if (this.master.canWrite) return true
  //if (this.master.canRead || !this.master.eof) return true //|| this.master.canWrite
  //this.master.close()
  //this.master = null
 }
 let symbols = '-*/\|'
 write("Connecting to ",this.address,symbols[Math.floor(Math.random()*symbols.length)],'\r')
 try{
  this.master = new Stream('tcp://' + this.address,0)
 } catch(err) 
 {
  debug(err)
  return false
 }
 this.master.writeln(this.name,' ',this.password)
 this.describe()
 writeln("\rConnected to  ",this.master,"  ")
 return true
}

Worker.prototype.describe = function()
{
 if (!this.master) return
 let msg = ''
 for (let n in this.handlers)
   msg += ' ' + n
 this.master.writeln('ACCEPT',msg)
}

Worker.prototype.addHandler = function(name, fn)
{
 this.handlers[name] = fn
 this.describe()
}

Worker.prototype.share = function(name,data)
{
  if (data.constructor.name == 'String')
  {
   this.master.writeln('DATA ',name,' ',data.length)
   this.master.write(data)  
  }
  else if (data.constructor.name == 'Number')
  {
   let s = data.toString()
   this.master.writeln('DATA ',name,' ',s.length)
   this.master.write(s)  
  }
  else if (data.constructor.name == 'Stream')
  {
   data.rewind()
   this.master.writeln('DATA ',name,' ',data.size())
   this.master.append(data)   
  }
  else
  {
   let s = data.toSource()
   this.master.writeln('DATA ',name,' ',s.length)
   this.master.write(s)
  }
}
Worker.prototype.alert = function(text)
{
 text = text.toString()
 this.master.writeln('ALERT ',text.length)
 this.master.write(text)
}

Worker.prototype.send = function(name,text)
{
 text = text.toString()
 this.master.writeln('SEND ',name,' ',text.length)
 this.master.write(text)
}

Worker.prototype.sendTo = function(host,name,text)
{
 text = text.toString()
 this.master.writeln('SENDTO ',host,' ',name,' ',text.length)
 this.master.write(text)
}

Worker.prototype.postTo = function(host,name,text)
{
 text = text.toString()
 this.master.writeln('POSTTO ',host,' ',name,' ',text.length)
 this.master.write(text)
}

Worker.prototype.resource = function(name,stream)
{
debug('GET ',name)
  let s = this.data[name]
  if (s) return s;
  this.master.writeln('GET ',name)
 
  while (!s)
  {
    if (this.master && this.master.wait(1000))
       s = this.step(name,stream) 
    if (!this.running) return null
    //system.sleep(10)
  }
  debug('Done GET ',name)
  s.rewind()
  return s
}

Worker.prototype.toStream = function() {return this.master;}

/*
 Single step the message loop, possibly waiting for a specific resource
*/
Worker.prototype.step = function(data_name,data_stream)
{
 if (!this.running) return false
 if (this.read())
 {
    let line = this.command
    this.reset()
       //let line = this.master.readln()
       debug('>',line)
       if (!line) return true;
       let [, message, text] = line.match(/(\S+)\s*(.*)/)
       switch (message.toUpperCase())
       {
         case 'STATUS':
          if (this.busy)      
            this.master.write('RUNNING ',this.job,' ',this.parameters.length,'\n',this.parameters)        
          else
            this.master.writeln('IDLE')
          break
         case 'DATA': //DATA name length
          let [name,length] = text.split(' ')
          if (data_name && name == data_name)
          {
            let s = data_stream || new Stream
            let l = parseInt(length)
            if (Number(length)) s.append(this.master,l)
            else debug('Bad number:',length)
            return s
          }
          else
          {
            let s = new Stream        
            this.data[name] = s
            if (length != '0')
            {
              s.append(this.master,Number(length))
              s.rewind()
            }
          }
          break
         case 'RUN': //RUN name length sender-id\nparameters
          {
            let [name,length,sender] = text.split(' ')
            let params = this.master.read(parseInt(length))
            this.sender = sender || 0
            if (this.busy)
            {
                this.master.writeln("BUSY ", name)
            }
            else
            {
                debug('running ',name,' ',params)
                let start = new Date
                this.busy = true
                this.job = name
                this.parameters = params

                if (this.handlers[name])
                {
                    this.master.writeln("RUNNING ", name)
                    try 
                    {
                        let result = this.handlers[name](params, this)
                        if (typeof result == "undefined") result = ''
                        else result = result.toString()
                        debug('done ',result)
                        let finish = new Date                 
                        this.master.writeln('RESULT ',finish-start,' ',result.length)
                        if (result.length) this.master.write(result)
                        debug(result)
                    } catch(x)
                    {
                        let msg = ''
                        for (let [n,v] in x) msg += ' ' + n + ': ' + v + ';'
                        this.master.writeln('ERROR ',name,msg)
                        debug('ERROR ',name,msg)
                    }
                }
                else this.master.writeln('ERROR ',name,' not understood')
                this.busy = false
            }
         }
        break
        case 'QUIT':
        debug('quitting')
         this.running = false;
         if (data_name) throw "Shutdown"
         return false
         break
        case 'ALERT':
         writeln(text)
         break
        case '':
         break
        default:
         throw "Invalid message: " +line
       }          
    // debug("Done message loop")
 }
 else if (this.master && this.master.canWrite)
 {
  //if (this.waiter++ == 10)
  {
   this.master.writeln()
  // this.waiter=0
  }
  //system.sleep(100)
 }
 else if (this.running)
 {
 let i=0
      while (!this.connect()) //wait for a connection
       {
         spin('connect')
         system.sleep(100 * (i++))
         if (i > 10) break;
      }
 }
 if (data_name) return null //GET loop, return a stream
 return this.running //normal loop, true means keep running
}

// stopper returns false to stop
Worker.prototype.run = function(stopper)
{
 let r = true 
 while (r && this.running && (stopper ? stopper() : true))
 {
  //spin('run')
  r = this.step()
  system.gc()
 }
 if (!this.running) 
 {
   this.master.writeln('BYE');
   this.master.read(1);
   this.master.close(); 
   this.master =null;
 }
}

Worker.join = function(address,password,name)
{
     let grid = new Worker(address,password,name)
     //run, load, and stop message handlers are included by default
     system.grid = grid
     if (!system._resource) system._resource = system.resource
     system.resource = function(name) {return system.grid.resource(name) || system._resource(name)}
     return grid;
}
